/*
 * Tapioca library
 * Copyright (C) 2006 INdT.
 * @author  Luiz Augusto von Dentz <luiz.dentz@indt.org.br>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "tpa-parameters.h"

#define DEBUG_DOMAIN TPA_DOMAIN_CONNECTION_MANAGER

#include <tapioca/base/tpa-debug.h>

struct _TpaParameter {
    gchar *name;
    TpaParameterFlags flags;
    GValue *value;
    GValue *default_value;
    void (* convert_func) (GValue *value, gchar **string_value);
};

struct _TpaParametersPrivate {
    GPtrArray *names;
    GHashTable *parameters;
    gboolean disposed;
};

G_DEFINE_TYPE(TpaParameters, tpa_parameters, G_TYPE_OBJECT)

static GType
tpa_parameters_gsignature (const gchar *sig)
{
    /* TODO: Add All types */
    if (g_str_has_prefix (sig, "s"))
        return G_TYPE_STRING;
    else if (g_str_has_prefix (sig, "b"))
        return G_TYPE_BOOLEAN;
    else if (g_str_has_prefix (sig, "q") || g_str_has_prefix (sig, "u"))
        return G_TYPE_UINT;
    else if (g_str_has_prefix (sig, "i") || g_str_has_prefix (sig, "n"))
        return G_TYPE_UINT;
    else
        return G_TYPE_NONE;
}

static void
tpa_parameters_dispose (GObject *object)
{
    TpaParameters *self = TPA_PARAMETERS (object);

    if (self->priv->disposed)
       /* If dispose did already run, return. */
       return;

    /* Make sure dispose does not run twice. */
    self->priv->disposed = TRUE;

    /* Free private data */
    if (self->priv->names)
        g_ptr_array_free (self->priv->names, FALSE);

    if (self->priv->parameters)
        g_hash_table_destroy (self->priv->parameters);

    G_OBJECT_CLASS (tpa_parameters_parent_class)->dispose (object);
}

static void
tpa_parameters_class_init (TpaParametersClass *klass)
{
    GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
    tpa_parameters_parent_class = g_type_class_peek_parent (klass);

    g_type_class_add_private (klass, sizeof (TpaParametersPrivate));

    gobject_class->dispose = tpa_parameters_dispose;
}

static void
tpa_parameters_init (TpaParameters *self)
{
    self->priv = TPA_PARAMETERS_GET_PRIVATE (self);
    self->priv->disposed = FALSE;
    self->priv->parameters = g_hash_table_new (g_str_hash, g_str_equal);
    self->priv->names = g_ptr_array_new ();
}

static void
tpa_parameters_convert_string (GValue *value,
                               gchar **string_value)
{
    if (value && !*string_value)
        *string_value = g_value_dup_string (value);
    else if (value && *string_value) {
        g_value_set_string (value, g_strdup (*string_value));
    }
}

static void
tpa_parameters_convert_boolean (GValue *value,
                                gchar **string_value)
{
    if (value && !*string_value)
        *string_value = g_value_get_boolean (value) ? g_strdup ("true") :
            g_strdup ("false");
    else if (value && *string_value)
        g_value_set_boolean (value, (g_str_equal (*string_value, "true")));
}

static void
tpa_parameters_convert_int (GValue *value,
                            gchar **string_value)
{
    if (value && !*string_value) {
        *string_value = g_new (gchar, 256);
        *string_value = g_ascii_dtostr (*string_value, 256, g_value_get_int (value));
    }
    else if (value && *string_value)
        g_value_set_int (value, g_ascii_strtod (*string_value, NULL));
}

static void
tpa_parameters_convert_uint (GValue *value,
                             gchar **string_value)
{
    if (value && !*string_value) {
        *string_value = g_new (gchar, 256);
        *string_value = g_ascii_dtostr (*string_value, 256, g_value_get_uint (value));
    }
    else if (value && *string_value)
        g_value_set_uint (value, g_ascii_strtod (*string_value, NULL));
}

static TpaParameter *
tpa_parameters_get_parameter (TpaParameters *self,
                              const gchar *name)
{
    TpaParameter *parameter = NULL;

    g_assert (self);
    g_return_val_if_fail (name != NULL, NULL);

    parameter = g_hash_table_lookup (self->priv->parameters, name);

    return parameter;
}

/**
 * tpa_parameters_new:
 * @returns: #TpaParameters instance.
 *
 * Create new instance of #TpaParameters.
 */
TpaParameters *
tpa_parameters_new (void)
{
    TpaParameters *self = NULL;

    self = TPA_PARAMETERS (g_object_new (TPA_TYPE_PARAMETERS, NULL));

    return self;
}

/**
 * tpa_parameters_add:
 * @self: #TpaParameters instance.
 * @name: New Parameter name.
 * @signature: Parameter type in form of dbus signature.
 * @returns: FALSE if fail.
 *
 * Add parameter name, if a parameter with same
 * name was already added it will fail.
 *
 */
gboolean
tpa_parameters_add_signature (TpaParameters *self,
                              const gchar *name,
                              const gchar *signature)
{
    GType type = G_TYPE_NONE;

    VERBOSE ("(%p, %s, %s)", self, name, signature);

    g_return_val_if_fail (signature != NULL, FALSE);

    type = tpa_parameters_gsignature (signature);

    return tpa_parameters_add (self, name, type);
}

/**
 * tpa_parameters_add:
 * @self: #TpaParameters instance.
 * @name: New Parameter name.
 * @type: Parameter #GType.
 * @returns: FALSE if fail.
 *
 * Add parameter name, if a parameter with same
 * name was already added it will fail.
 *
 */
gboolean
tpa_parameters_add (TpaParameters *self,
                    const gchar *name,
                    GType type)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %s)", self, name, g_type_name (type));

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter == NULL, FALSE);

    parameter = g_new0 (TpaParameter, 1);
    parameter->name = g_strdup (name);
    parameter->flags = TPA_PARAMETER_FLAGS_NONE;
    parameter->default_value = g_new0 (GValue, 1);
    parameter->value = g_new0 (GValue, 1);
    g_value_init (parameter->default_value, type);
    g_value_init (parameter->value, type);

    switch (type) {
        case G_TYPE_STRING:
            parameter->convert_func = tpa_parameters_convert_string;
            break;
        case G_TYPE_BOOLEAN:
            parameter->convert_func = tpa_parameters_convert_boolean;
            break;
        case G_TYPE_UINT:
            parameter->convert_func = tpa_parameters_convert_uint;
            break;
        case G_TYPE_INT:
            parameter->convert_func = tpa_parameters_convert_int;
            break;
    }

    g_hash_table_insert (self->priv->parameters, parameter->name, parameter);
    g_ptr_array_add (self->priv->names, parameter->name);

    return TRUE;
}

/**
 * tpa_parameters_remove:
 * @self: #TpaParameters instance.
 * @name: New Parameter name.
 * @returns: FALSE if fail.
 *
 * Remove parameter if no parameter is found it
 * will fail.
 *
 */
gboolean
tpa_parameters_remove (TpaParameters *self,
                       const gchar *name)
{
    TpaParameter *parameter = NULL;
    guint i = 0;
    guint len = 0;

    VERBOSE ("(%p, %s)", self, name);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);

    g_hash_table_remove (self->priv->parameters, name);

    for (i = 0, len = self->priv->names->len; i < len; i++) {
        if (!g_str_equal (g_ptr_array_index (self->priv->names, i), name)) {
            g_ptr_array_remove_index (self->priv->names, i);
            break;
        }
    }

    return TRUE;
}

/**
 * tpa_parameters_get_all:
 * @self: #TpaParameters instance.
 * @returns: #GPtrArray with parameters names.
 *
 * Get all parameters names available.
 *
 */
GPtrArray *
tpa_parameters_get_all (TpaParameters *self)
{
    GPtrArray *parameters = NULL;
    guint i = 0;
    guint len = 0;

    VERBOSE ("(%p)", self);
    g_assert (self);

    parameters = g_ptr_array_new ();
    len = self->priv->names->len;
    for (i = 0; i < len; i++)
        g_ptr_array_add (parameters, g_ptr_array_index (self->priv->names, i));

    return parameters;
}

/**
 * tpa_parameters_get_required:
 * @self: #TpaParameters instance.
 * @returns: #GPtrArray with parameters names.
 *
 * Get all parameters names available that has required flag.
 *
 */
GPtrArray *
tpa_parameters_get_required (TpaParameters *self)
{
    GPtrArray *parameters = NULL;
    guint i = 0;

    VERBOSE ("(%p)", self);
    parameters = tpa_parameters_get_all (self);

    for (i = 0; i < parameters->len; i++) {
        gchar *parameter = g_ptr_array_index (parameters, i);
        if (tpa_parameters_has_required_flag (self, parameter)) {
            g_ptr_array_remove_index(parameters, i);
            --i;
        }
    }

    return parameters;
}

/**
 * tpa_parameters_get_register:
 * @self: #TpaParameters instance.
 * @returns: #GPtrArray with parameters names.
 *
 * Get all parameters names available that has register flag.
 *
 */
GPtrArray *
tpa_parameters_get_register (TpaParameters *self)
{
    GPtrArray *parameters = NULL;
    guint i = 0;

    VERBOSE ("(%p)", self);
    parameters = tpa_parameters_get_all (self);

    for (i = 0; i < parameters->len; i++) {
        gchar *parameter = g_ptr_array_index (parameters, i);
        if (tpa_parameters_has_register_flag (self, parameter)) {
            g_ptr_array_remove_index(parameters, i);
            --i;
        }
    }

    return parameters;
}

/**
 * tpa_parameters_get_default_value:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: NULL if fail.
 *
 * Get copy of parameter default value that must be free.
 *
 */
GValue *
tpa_parameters_get_default_value (TpaParameters *self,
                                  const gchar *name)
{
    TpaParameter *parameter = NULL;
    GValue *value = NULL;

    VERBOSE ("(%p, %s)", self, name);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, NULL);

    value = g_new0 (GValue, 1);
    g_value_init (value, G_VALUE_TYPE (parameter->default_value));
    g_value_copy (parameter->default_value, value);

    return value;
}

/**
 * tpa_parameters_get_default_value_as_string:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: NULL if fail.
 *
 * Function to get any parameter default value to string.
 * Note that self function convert parameter value to string.
 *
 * Supports convertions to:
 *   boolean:
 *     "true" for TRUE value.
 *     "false" for FALSE value.
 *   double
 *   integer
 *   string
 *
 */
gchar *
tpa_parameters_get_default_value_as_string (TpaParameters *self,
                                            const gchar *name)
{
    TpaParameter *parameter = NULL;
    gchar *value_as_string = NULL;

    VERBOSE ("(%p, %s)", self, name);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, NULL);

    parameter->convert_func (parameter->default_value, &value_as_string);

    return value_as_string;
}

/**
 * tpa_parameters_get_default_value_as_uint:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: 0 if fail.
 *
 * Fast function to get parameter default value that has known type as unsigned
 * integer.
 * DO NOT USE IT FOR NON-UINTEGER PARAMETERS.
 *
 */
guint
tpa_parameters_get_default_value_as_uint (TpaParameters *self,
                                          const gchar *name)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s)", self, name);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, 0);
    g_return_val_if_fail (G_VALUE_HOLDS_UINT (parameter->default_value), 0);

    return g_value_get_uint (parameter->default_value);
}

/**
 * tpa_parameters_get_default_value_as_int:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: 0 if fail.
 *
 * Fast function to get parameter default value that has known type as integer.
 * DO NOT USE IT FOR NON-INTEGER PARAMETERS.
 *
 */
gint
tpa_parameters_get_default_value_as_int (TpaParameters *self,
                                         const gchar *name)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s)", self, name);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, 0);
    g_return_val_if_fail (G_VALUE_HOLDS_INT (parameter->default_value), 0);

    return g_value_get_uint (parameter->default_value);
}

/**
 * tpa_parameters_get_default_value_as_boolean:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: -1 if fail.
 *
 * Fast function to get parameter default value that has known type as boolean.
 * DO NOT USE IT FOR NON-BOOLEAN PARAMETERS.
 *
 */
gboolean
tpa_parameters_get_default_value_as_boolean (TpaParameters *self,
                                             const gchar *name)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s)", self, name);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS_BOOLEAN (parameter->default_value), FALSE);

    return g_value_get_uint (parameter->default_value);
}

/**
 * tpa_parameters_set_default_value:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value: New GValue value.
 * @returns: FALSE if fail.
 *
 * Set parameter default value, givin value must hold the same type as parameter.
 *
 */
gboolean
tpa_parameters_set_default_value (TpaParameters *self,
                                  const gchar *name,
                                  GValue *value)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %p)", self, name, value);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS (value, G_VALUE_TYPE (parameter->default_value)),
         FALSE);

    g_value_copy (value, parameter->default_value);
    g_value_copy (value, parameter->value);

    return TRUE;
}

/**
 * tpa_parameters_set_default_value_as_string:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value_as_string: New string value or string to be converted.
 * @returns: FALSE if fail.
 *
 * Function to set any string default value to parameter value.
 * Note that self function convert the given value to
 * parameter type if it is not string type.
 *
 * Supports convertions to:
 *   boolean:
 *     "true" for TRUE value.
 *     "false" for FALSE value.
 *   double
 *   integer
 *   string
 */
gboolean
tpa_parameters_set_default_value_as_string (TpaParameters *self,
                                            const gchar *name,
                                            const gchar *value_as_string)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %s)", self, name, value_as_string);
    g_return_val_if_fail (value_as_string != NULL, FALSE);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);

    parameter->convert_func (parameter->default_value, (gchar **) &value_as_string);

    return TRUE;
}

/**
 * tpa_parameters_set_default_value_as_uint:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value_as_uint: New integer value.
 * @returns: FALSE if fail.
 *
 * Fast function to set parameter default value that has known type as unsigned integer.
 * DO NOT USE IT FOR NON-UINTEGER PARAMETERS.
 *
 */
gboolean
tpa_parameters_set_default_value_as_uint (TpaParameters *self,
                                          const gchar *name,
                                          guint value_as_uint)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %d)", self, name, value_as_uint);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS_UINT (parameter->default_value), FALSE);

    g_value_set_uint (parameter->default_value, value_as_uint);

    return TRUE;
}

/**
 * tpa_parameters_set_default_value_as_int:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value_as_int: New integer value.
 * @returns: FALSE if fail.
 *
 * Fast function to set parameter default value that has known type as integer.
 * DO NOT USE IT FOR NON-INTEGER PARAMETERS.
 *
 */
gboolean
tpa_parameters_set_default_value_as_int (TpaParameters *self,
	                                     const gchar *name,
	                                     gint value_as_int)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %d)", self, name, value_as_int);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS_INT (parameter->default_value), FALSE);

    g_value_set_int (parameter->default_value, value_as_int);

    return TRUE;
}

/**
 * tpa_parameters_set_default_value_as_boolean:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value_as_boolean: New boolean value.
 * @returns: FALSE if fail.
 *
 * Fast function to set parameter dafault value that has known type as boolean.
 * DO NOT USE IT FOR NON-BOOLEAN PARAMETERS.
 *
 */
gboolean
tpa_parameters_set_default_value_as_boolean (TpaParameters *self,
                                             const gchar *name,
                                             gboolean value_as_boolean)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %s)", self, name, value_as_boolean ? "true" : "false");

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS_BOOLEAN (parameter->default_value), FALSE);

    g_value_set_boolean (parameter->default_value, value_as_boolean);

    return TRUE;
}

/**
 * tpa_parameters_get_value:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: NULL if fail.
 *
 * Get copy of parameter value that must be free.
 *
 */
GValue *
tpa_parameters_get_value (TpaParameters *self,
                          const gchar *name)
{
    TpaParameter *parameter = NULL;
    GValue *value = NULL;

    VERBOSE ("(%p, %s)", self, name);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, NULL);

    value = g_new0 (GValue, 1);
    g_value_init (value, G_VALUE_TYPE (parameter->value));
    g_value_copy (parameter->value, value);

    return value;
}

/**
 * tpa_parameters_get_value_as_string:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: NULL if fail.
 *
 * Function to get any parameter value to string.
 * Note that self function convert parameter value to string.
 *
 * Supports convertions to:
 *   boolean:
 *     "true" for TRUE value.
 *     "false" for FALSE value.
 *   double
 *   integer
 *   string
 *
 */
gchar *
tpa_parameters_get_value_as_string (TpaParameters *self,
                                    const gchar *name)
{
    TpaParameter *parameter = NULL;
    gchar *value_as_string = NULL;

    VERBOSE ("(%p, %s)", self, name);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, NULL);

    /* Call callback to convert value to string */
    parameter->convert_func (parameter->value, &value_as_string);

    return value_as_string;
}

/**
 * tpa_parameters_get_value_as_uint:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: 0 if fail.
 *
 * Fast function to get parameter value that has known type as unsigned
 * integer.
 * DO NOT USE IT FOR NON-UINTEGER PARAMETERS.
 *
 */
guint
tpa_parameters_get_value_as_uint (TpaParameters *self,
                                  const gchar *name)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s)", self, name);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, 0);
    g_return_val_if_fail (G_VALUE_HOLDS_UINT (parameter->value), 0);

    return g_value_get_uint (parameter->value);
}

/**
 * tpa_parameters_get_value_as_int:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: 0 if fail.
 *
 * Fast function to get parameter value that has known type as integer.
 * DO NOT USE IT FOR NON-INTEGER PARAMETERS.
 *
 */
gint
tpa_parameters_get_value_as_int (TpaParameters *self,
                                 const gchar *name)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s)", self, name);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, 0);
    g_return_val_if_fail (G_VALUE_HOLDS_INT (parameter->value), 0);

    return g_value_get_int (parameter->value);
}

/**
 * tpa_parameters_get_value_as_boolean:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: -1 if fail.
 *
 * Fast function to get parameter value that has known type as boolean.
 * DO NOT USE IT FOR NON-BOOLEAN PARAMETERS.
 *
 */
gboolean
tpa_parameters_get_value_as_boolean (TpaParameters *self,
                                     const gchar *name)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s)", self, name);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, -1);
    g_return_val_if_fail (G_VALUE_HOLDS_BOOLEAN (parameter->value), -1);

    return g_value_get_boolean (parameter->value);
}

/**
 * tpa_parameters_set_value:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value: New GValue value.
 * @returns: FALSE if fail.
 *
 * Set parameter value, givin value must hold the same type as parameter.
 *
 */
gboolean
tpa_parameters_set_value (TpaParameters *self,
                          const gchar *name,
                          GValue *value)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %p)", self, name, value);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS (value, G_VALUE_TYPE (parameter->value)),
         FALSE);

    g_value_copy (value, parameter->value);

    return TRUE;
}

/**
 * tpa_parameters_set_value_as_string:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value_as_string: New string value or string to be converted.
 * @returns: FALSE if fail.
 *
 * Function to set any string value to parameter value.
 * Note that this function convert the given value to
 * parameter type if it is not string type.
 *
 * Supports convertions to:
 *   boolean:
 *     "true" for TRUE value
 *     "false" for FALSE value
 *   double
 *   integer
 *   string
 */
gboolean
tpa_parameters_set_value_as_string (TpaParameters *self,
                                    const gchar *name,
                                    const gchar *value_as_string)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %s)", self, name, value_as_string);
    g_return_val_if_fail (value_as_string != NULL, FALSE);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);

    /* Call callback to convert string to value */
    parameter->convert_func (parameter->value, (gchar **) &value_as_string);

    return TRUE;
}

/**
 * tpa_parameters_set_value_as_uint:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value_as_uint: New unsigned integer value.
 * @returns: FALSE if fail.
 *
 * Fast function to set parameter value that has known type as unsigned
 * integer.
 * DO NOT USE IT FOR NON-UINTEGER PARAMETERS.
 *
 */
gboolean
tpa_parameters_set_value_as_uint (TpaParameters *self,
                                  const gchar *name,
                                  guint value_as_uint)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %d)", self, name, value_as_uint);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS_UINT (parameter->value), FALSE);

    g_value_set_uint (parameter->value, value_as_uint);

    return TRUE;
}

/**
 * tpa_parameters_set_value_as_int:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value_as_int: New integer value.
 * @returns: FALSE if fail.
 *
 * Fast function to set parameter value that has known type as integer.
 * DO NOT USE IT FOR NON-INTEGER PARAMETERS.
 *
 */
gboolean
tpa_parameters_set_value_as_int (TpaParameters *self,
                                 const gchar *name,
                                 gint value_as_int)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %d)", self, name, value_as_int);

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS_INT (parameter->value), FALSE);

    g_value_set_int (parameter->value, value_as_int);

    return TRUE;
}

/**
 * tpa_parameters_set_value_as_boolean:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @value_as_boolean: New boolean value.
 * @returns: FALSE if fail.
 *
 * Fast function to set parameter value that has known type as boolean.
 * DO NOT USE IT FOR NON-BOOLEAN PARAMETERS.
 *
 */
gboolean
tpa_parameters_set_value_as_boolean (TpaParameters *self,
                                     const gchar *name,
                                     gboolean value_as_boolean)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %s)", self, name, value_as_boolean ? "true" : "false");

    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);
    g_return_val_if_fail (G_VALUE_HOLDS_BOOLEAN (parameter->value), FALSE);

    g_value_set_boolean (parameter->value, value_as_boolean);

    return TRUE;
}

/**
 * tpa_parameters_reset:
 * @self: #TpaParameters instance
 *
 * Reset all parameters to its default values.
 * Note that this operation discard old parameter value.
 *
 */
void
tpa_parameters_reset (TpaParameters *self)
{
    GPtrArray *parameters = NULL;
    guint i = 0;

    VERBOSE ("(%p)", self);
    parameters = tpa_parameters_get_all (self);

    for (i = 0; i < parameters->len; i++) {
        TpaParameter *parameter =
            tpa_parameters_get_parameter (self, g_ptr_array_index (parameters, i));

        g_value_copy (parameter->default_value, parameter->value);
    }
}

/**
 * tpa_parameters_get_flags:
 * @self: #TpaParameters instance
 * @name: Parameter name.
 * @returns: #TpaParameterFlags
 *
 * Get parameter flags.
 *
 */
TpaParameterFlags
tpa_parameters_get_flags (TpaParameters *self,
                          const gchar *name)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s)", self, name);
    parameter = tpa_parameters_get_parameter (self, name);

    if (!parameter)
        return TPA_PARAMETER_FLAGS_NONE;

    return parameter->flags;
}

/**
 * tpa_parameters_set_flags:
 * @self: #TpaParameters instance
 * @returns: FALSE if fail.
 *
 * Set parameter flags.
 *
 */
gboolean
tpa_parameters_set_flags (TpaParameters *self,
                          const gchar *name,
                          TpaParameterFlags flags)
{
    TpaParameter *parameter = NULL;

    VERBOSE ("(%p, %s, %d)", self, name, flags);
    parameter = tpa_parameters_get_parameter (self, name);

    g_return_val_if_fail (parameter != NULL, FALSE);

    parameter->flags |= flags;

    return TRUE;
}

/**
 * tpa_parameters_get_formatted_table:
 * @self: #TpaParameters instance
 * @returns: formatted parameters into a GHashTable or NULL if
 * fail.
 *
 * Function to format parameters into a new GHashTable to be
 * sent to connection managers.
 *
 */
GHashTable *
tpa_parameters_get_formatted_table (TpaParameters *self)
{
    GPtrArray *parameters = NULL;
    GHashTable *table = NULL;
    guint i = 0;

    VERBOSE ("(%p)", self);
    parameters = tpa_parameters_get_all (self);
    if (parameters) {
        table = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
        for (i = 0; i < parameters->len; i++) {
            gchar *name = g_ptr_array_index (parameters, i);
            gchar *value = tpa_parameters_get_value_as_string (self, name);
            gchar *default_value = tpa_parameters_get_default_value_as_string (self, name);
            /* Add parameters that has value and does not match with default value*/
            if (value && !g_str_equal (value, "") && !g_str_equal (value, "0") &&
                (!default_value || !g_str_equal (value, default_value))) {
                GValue *gvalue = tpa_parameters_get_value (self, name);
                g_hash_table_insert (table, name, gvalue);
                if ((g_str_equal ("password", name)) || (g_str_equal ("passwd", name))) {
                    INFO ("parameter %s value %s added", name, "*******");
                }
                else {
                    INFO ("parameter %s value %s added", name, value);
                }
            }
        }
    }

    return table;
}
